#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Backup."""

from __future__ import with_statement

import os
import re
import shutil

from iobase.xmldata import ANY_CHILD_PATH
from iobase.xmldata import (XmlNode, ANY_CHILD_NAME)
from iobase.exception import NotFoundError
from ionas.exception import NoEntryError, InternalError
import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)
from albg_info import ALBG_BACKUPS_PATH
from libpal.crond import ctrl as crond
import pal_config as config
from pal_config import (DATE_CMD)
import ionas.fileaccess
from ionas.fileaccess import SmbAccess, LocalAccess
from ionas.request import (Entry, Request, RequestData,)

import fal_info

# For cloud packages
try:
    from libalbg.backup.access.fletsazukeru import (BackupFletsAzukeruAccess,)
    from libalbg.backup.access.fletsazukerutree import (FletsAzukeruFileTree,)
    from landiskcloud.info.fletsazukeru import (FletsAzukeruConnectionInfo,
                                                FletsAzukeruTargetInfo)
except:
    pass

try:
    from libalbg.backup.access.amazons3 import (BackupAmazonS3Access,)
    from libalbg.backup.access.amazons3tree import (AmazonS3FileTree,)
    from landiskcloud.info.amazons3 import (AmazonS3ConnectionInfo,
                                            AmazonS3TargetInfo)
except:
    pass

try:
    from libalbg.backup.access.nifcloud import (BackupNifCloudAccess,)
    from libalbg.backup.access.nifcloudtree import (NifCloudFileTree,)
    from landiskcloud.info.nifcloud import (NifCloudConnectionInfo,
                                            NifCloudTargetInfo)
except:
    pass

try:
    from libalbg.backup.access.azureaccess import (BackupAzureAccess,)
    from libalbg.backup.access.azuretree import (AzureFileTree,)
    from landiskcloud.info.azure import (AzureConnectionInfo,
                                            AzureTargetInfo)
except:
    pass

try:
    from libalbg.backup.access.googledriveaccess import (
                                             BackupGoogleDriveAccess,)
    from libalbg.backup.access.googledrivetree import (GoogleDriveFileTree,)
    from landiskcloud.info.googledrive import (GoogleDriveConnectionInfo,
                                            GoogleDriveTargetInfo)
except:
    pass

try:
    from libalbg.backup.access.dropboxaccess import (
                                             BackupDropboxAccess,)
    from libalbg.backup.access.dropboxtree import (DropboxFileTree,)
    from landiskcloud.info.dropbox import (DropboxConnectionInfo,
                                            DropboxTargetInfo)
except:
    pass

try:
    from libalbg.backup.access.boxaccess import (
                                             BackupBoxAccess,)
    from libalbg.backup.access.boxtree import (BoxFileTree,)
    from landiskcloud.info.box import (BoxConnectionInfo, BoxTargetInfo)
except:
    pass

try:
    from libalbg.backup.access.mdcloud import (BackupMdCloudAccess,)
    from libalbg.backup.access.mdcloudtree import (MdCloudFileTree,)
    from landiskcloud.info.mdcloud import (MdCloudConnectionInfo,
                                            MdCloudTargetInfo)
except:
    pass

try:
    from libalbg.backup.access.workingfolder import BackupWorkingFolderAccess
    from libalbg.backup.access.workingfoldertree import WorkingFolderFileTree
    from landiskcloud.info.workingfolder import WorkingFolderTargetInfo
except:
    pass

try:
    from libalbg.backup.access.homebox2 import BackupHomeBox2Access
    from libalbg.backup.access.homebox2tree import HomeBox2FileTree
    from landiskcloud.info.homebox2 import HomeBox2TargetInfo
except:
    pass

try:
    from landiskcloud.proxy import (ProxyInfo,)
except:
    pass


DAY_NAME = "day"
ID_NAME = "id"
HOUR_PATH = "time/hour"
MINUTE_PATH = "time/minute"
SCHEDULE_PATH = "job/schedule"
DATE_PATH = "job/schedule/date"
NAME_PATH = "//job/name"
ID_PATH = "//job/id"

FMT_ENTRYNAME = "%s_%s"
FMT_CMD_ARGS = '/usr/local/bin/exec_backup %s "%s"'

FMT_ADD_DATE_CMD_ARGS = \
   '[ `' + DATE_CMD + ' +\\%%Y` -eq %s ] && /usr/local/bin/exec_backup %s "%s"'


COMPLETE_STATUS = "COMPLETE"
NOT_EXEC_STATUS = "NOT_EXEC"

DISABLE_REMAIN_TIMES = "0"

OPTION_NAME = "option"
NAME_NAME = "name"
GENERATIONLIST_NAME = "generationlist"
GENERATION_NAME = "generation"
LOCALHOST_NAME = "localhost"
JOB_NAME = "job"
DESTINATION_NAME = "destination"
DETAIL_ATTR = "detail"
TIMES_ATTR = "times"
PATH_NAME = "path"
TYPE_NAME = "type"
SHARENAME_NAME = "sharename"
SUBDIR_NAME = "subdir"
UNC_NAME = "unc"
USERNAME_NAME = "username"
PASSWORD_NAME = "password"

YEAR_NAME = "year"
MONTH_NAME = "month"
HOUR_NAME = "hour"
MINUTE_NAME = "minute"

UNC_RE = r"^\\\\[^\\]+\\.+"
UNC_HOST_RE = r"^\\\\([^\\]+)\\"
UNC_SHARE_RE = r"^\\\\[^\\]+\\(.+)"
GENERATION_ID_RE = "(^[0-9]{14})"

SMB_PREFIX = "smb://"
SMB_SEP = "/"

UNC_SEP = "\\"
VOL_SEP = "/"

LOCAL_TYPE = "local"
SMB_TYPE = "smb"
FLETSAZUKERU_TYPE = "fletsazukeru"
AMAZONS3_TYPE = "amazons3"
NIFCLOUD_TYPE = "nifcloud"
AZURE_TYPE = "azure"
GOOGLEDRIVE_TYPE = "googledrive"
DROPBOX_TYPE = "dropbox"
BOX_TYPE = "box"
MDCLOUD_TYPE = "mdcloud"
WORKINGFOLDER_TYPE = "workingfolder"
HOMEBOX2_TYPE = "homebox2"
ALL_LOCAL_TYPE = "all_local"

CONNECTION_ID_PATH = "connection/id"
FIND_BY_ID_XPATH_TMPL = ANY_CHILD_NAME + '/%s/id[text()="%s"]/..'

URL_NAME = "url"
HOST_NAME_PATH = "hostdesc/name"


def get_proxy(url):
    """Try to get proxy and return it."""

    entry = Entry(ReqTypes.QUERY,
                    fal_info.FAL_PROXY_REQ_BASE_XML[ReqTypes.QUERY])

    entry.get().get_node(URL_NAME).set_text(url)
    request_data = RequestData(entry)
    target = Request(fal_info.FAL_PROXY_PATH)
    result_entry = target.request(request_data).get_entrylist().get_entry()
    if result_entry.get_result() != Results.OK:
        return None

    try:
        return ProxyInfo(result_entry.get()).get_proxy_url("https")
    except:
        return None


def get_proxy_host_port(url):
    """Try to get proxy and return it."""

    entry = Entry(ReqTypes.QUERY,
                    fal_info.FAL_PROXY_REQ_BASE_XML[ReqTypes.QUERY])

    entry.get().get_node(URL_NAME).set_text(url)
    request_data = RequestData(entry)
    target = Request(fal_info.FAL_PROXY_PATH)
    result_entry = target.request(request_data).get_entrylist().get_entry()
    if result_entry.get_result() != Results.OK:
        return (None, None)

    try:
        proxy_info = ProxyInfo(result_entry.get())
        return (proxy_info.host, proxy_info.port)
    except:
        return (None, None)


def get_tip_node(node, fullpath):

    for path in fullpath.split("/"):
        try:
            node.get_node(path)
        except NotFoundError:
            node.append_child(path)
        finally:
            node = node.get_node(path)
    return node


def get_local_host_name():

    sysinfo_config = Request(fal_info.FAL_SYSINFO_PATH,
        filter_cmd=config.SUDO_CMD).request(
        Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()
    return sysinfo_config.get_node(HOST_NAME_PATH).get_text()


class PalBackups(palcmd.PalCmd):
    """Pal Backup command implementation class."""

    def __add_schedule_entry(self, category, job_name, job_id, schedule_node):

        if not schedule_node.is_enabled():
            return

        week = ",".join([
            node.get_text() for node in schedule_node.get_node_list(DAY_NAME)])
        hour = schedule_node.get_node(HOUR_PATH).get_text()
        minutes = schedule_node.get_node(MINUTE_PATH).get_text()

        crond.add_entry(
            FMT_ENTRYNAME % (category, job_id),
            FMT_CMD_ARGS % (category, job_name),
            user="root",
            week=week,
            hour=hour,
            minutes=minutes)

    def __add_date_entry(self, category, job_name, job_id, date_node):

        year = date_node.get_node(YEAR_NAME).get_text()
        month = date_node.get_node(MONTH_NAME).get_text()
        day = date_node.get_node(DAY_NAME).get_text()
        hour = date_node.get_node(HOUR_NAME).get_text()
        minutes = date_node.get_node(MINUTE_NAME).get_text()

        crond.add_entry(
            FMT_ENTRYNAME % (category, job_id),
            FMT_ADD_DATE_CMD_ARGS % (year, category, job_name),
            user="root",
            minutes=minutes,
            hour=hour,
            day=day,
            month=month)

    def __add_cron_entry(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        job_name = body.get_node(NAME_PATH).get_text(False)
        job_id = body.get_node(ID_PATH).get_text()

        try:
            self.__add_schedule_entry(
                category, job_name, job_id,
                body.get_node(category).get_node(SCHEDULE_PATH))
        except NotFoundError:
            pass

        try:
            self.__add_date_entry(
                category, job_name, job_id,
                body.get_node(category).get_node(DATE_PATH))
        except NotFoundError:
            pass

        entry.set_result(Results.OK)
        return entry

    def __delete_cron_entry(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        job_id = body.get_node(ID_PATH).get_text()

        crond.delete_entry(FMT_ENTRYNAME % (category, job_id))

        entry.set_result(Results.OK)
        return entry

    def __get_file_info(self, file_name):

        file_path = os.path.join(config.BACKUP_STATUS_DIR, file_name)
        with open(file_path, "rb") as f:
            info_list = f.read().split(":", 1)
            if len(info_list) == 2:
                status, result = info_list
            else:
                status = info_list[0]
                result = ""

        timestamp = str(int(os.stat(file_path).st_mtime))

        return (status, result, timestamp)

    def __set_status(self, category, id, status):

        file_name = category + "_" + id
        tmp_file_path = os.path.join(
            config.TEMP_DIR, file_name + "." + str(os.getpid()))
        file_path = os.path.join(
            config.BACKUP_STATUS_DIR, file_name)

        with open(tmp_file_path, "wb") as tmp_file:
            tmp_file.write(status)

        shutil.move(tmp_file_path, file_path)

    def __set_remain_times(self, category, id, times):

        file_name = category + "_" + id
        tmp_file_path = os.path.join(
            config.TEMP_DIR, file_name + "." + str(os.getpid()))
        file_path = os.path.join(
            config.BACKUP_REMAIN_TIMES_DIR, file_name)

        with open(tmp_file_path, "wb") as tmp_file:
            tmp_file.write(str(times))

        shutil.move(tmp_file_path, file_path)

    def __set_status_to_xml(self, status_fname, entry):

        status, result, timestamp = self.__get_file_info(status_fname)
        category, job_id = status_fname.split("_")

        joblist_node = get_tip_node(entry.get(), category + "/joblist")
        joblist_node.append_child("job")
        job_node = joblist_node.get_node_list("job")[-1]

        job_node.append_child("id", job_id)
        job_node.append_child("status", status)

        if status == COMPLETE_STATUS:
            job_node.append_child("result", result)
            job_node.append_child("timestamp", timestamp)

    def __set_xml_to_status(self, category, job_node):

        id = job_node.get_value("id")
        status = job_node.get_value("status")

        self.__set_status(category, id, status)

    def __set_accesser(self, accesser):

        self.dst_listdir = lambda path:\
            accesser.listdir(path)

    def __set_remote_accesser(self, username, passwd):

        return self.__set_accesser(SmbAccess(username, passwd))

    def __set_local_accesser(self):

        return self.__set_accesser(LocalAccess())

    def __get_remote_abspath(self, unc, job_name):

        share_path = \
            re.match(UNC_SHARE_RE, unc).group(1).replace(UNC_SEP, SMB_SEP)
        host = re.match(UNC_HOST_RE, unc).group(1)
        return SMB_PREFIX + SMB_SEP.join([host, share_path, job_name])

    def __get_local_abspath(self, basepath, subdir, job_name):

        sanitized_subdir = re.sub(r'[\\/]', os.sep, subdir)
        if sanitized_subdir.startswith(os.sep):
            raise InternalError(
                detail="Invalid subdir (absolute path)", path=subdir)

        abspath = os.path.join(basepath, sanitized_subdir, job_name)
        normalized_abspath = os.path.normpath(abspath)
        if abspath != normalized_abspath:
            raise InternalError(
                detail="Invalid subdir (not normalized)", path=subdir)

        return normalized_abspath

    def _get_remote_dst_user_info(self, node):

        try:
            username = node.get_node(USERNAME_NAME).get_text()
        except NotFoundError:
            username = ""

        try:
            passwd = node.get_node(PASSWORD_NAME).get_text()
        except NotFoundError:
            passwd = ""

        return (username, passwd)

    def __get_generation_dir_list(self, abspath):
        """Get all generation list."""

        gen_dir_list = []
        try:
            for gen_dir in self.dst_listdir(abspath):
                match = re.match(GENERATION_ID_RE, gen_dir)
                if match:
                    gen_dir_list.append(match.group(1))
        except NoEntryError as e:
            self.warning("Failed to access path: " + e.path)
            raise

        return gen_dir_list

    def __get_generation_list_node(self, opt_job_node):

        job_name = opt_job_node.get_node(NAME_NAME).get_text()
        dst_node = opt_job_node.get_node(DESTINATION_NAME)
        dst_type = dst_node.get_value(TYPE_NAME, "")

        if dst_type == SMB_TYPE:
            share = dst_node.get_value(SHARENAME_NAME, "")
            subdir = dst_node.get_value(SUBDIR_NAME, "")
            unc = dst_node.get_value(UNC_NAME, "")
            if re.match(UNC_RE, unc):
                share = re.match(UNC_SHARE_RE, unc).group(1)
                host = re.match(UNC_HOST_RE, unc).group(1)
                username, passwd = \
                    self._get_remote_dst_user_info(
                        opt_job_node.get_node(DESTINATION_NAME))
                abspath = self.__get_remote_abspath(unc, job_name)
                self.__set_remote_accesser(username, passwd)
        elif dst_type == LOCAL_TYPE or dst_type == ALL_LOCAL_TYPE:
            # UNC形式でなければローカル共有名として処理
            if dst_type == ALL_LOCAL_TYPE:
                host = get_local_host_name()
            else:
                host = LOCALHOST_NAME
            share = dst_node.get_value(SHARENAME_NAME, "")
            subdir = dst_node.get_value(SUBDIR_NAME, "")
            unc = dst_node.get_value(UNC_NAME, "")
            if subdir and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = "\\\\%s\\%s%s" % (host, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))

            username, passwd = ("", "")
            abspath = self.__get_local_abspath(
                dst_node.get_node(PATH_NAME).get_text(),
                                    subdir.lstrip(VOL_SEP), job_name)
            self.__set_local_accesser()
        elif dst_type == FLETSAZUKERU_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (FLETSAZUKERU_TYPE, target_id,))

            target_info = FletsAzukeruTargetInfo(dst_node, target_node)

            proxy = None
            try:
                proxy = get_proxy("https://%s" %
                            (target_info.connection_info.host_name,))
            except:
                proxy = None

            host = dst_type
            share = target_info.connection_info.name
            subdir = target_info.subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP

            abspath += VOL_SEP + job_name
            self.__set_accesser(BackupFletsAzukeruAccess(target_info,
                                                            proxy))
        elif dst_type == AMAZONS3_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (AMAZONS3_TYPE, target_id,))

            target_info = AmazonS3TargetInfo(dst_node, target_node)

            try:
                proxy_host, proxy_port = get_proxy_host_port(
                            "https://%s" %
                                (target_info.connection_info.endpoint,))
            except:
                proxy_host = None
                proxy_port = None

            host = dst_type
            share = target_info.connection_info.name + "/" + \
                                target_info.bucket
            subdir = target_info.subdir
            subdir = subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if  abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP

            abspath += VOL_SEP + job_name
            self.__set_accesser(BackupAmazonS3Access(target_info,
                                        proxy_host, proxy_port))
        elif dst_type == NIFCLOUD_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (NIFCLOUD_TYPE, target_id,))

            target_info = NifCloudTargetInfo(dst_node, target_node)

            try:
                proxy_host, proxy_port = get_proxy_host_port(
                            "https://%s" %
                                (target_info.connection_info.endpoint,))
            except:
                proxy_host = None
                proxy_port = None

            host = dst_type
            share = target_info.connection_info.name + "/" + \
                                target_info.bucket
            subdir = target_info.subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if  abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP

            abspath += VOL_SEP + job_name
            self.__set_accesser(BackupNifCloudAccess(target_info,
                                        proxy_host, proxy_port))
        elif dst_type == AZURE_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (AZURE_TYPE, target_id,))

            target_info = AzureTargetInfo(dst_node, target_node)

            try:
                proxy_host, proxy_port = get_proxy_host_port(
                            "https://www.microsoft.com/")
            except:
                proxy_host = None
                proxy_port = None

            host = dst_type
            share = target_info.connection_info.name + "/" + \
                                target_info.container
            subdir = target_info.subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if  abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP

            abspath += VOL_SEP + job_name
            self.__set_accesser(BackupAzureAccess(target_info,
                                        proxy_host, proxy_port))
        elif dst_type == GOOGLEDRIVE_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (GOOGLEDRIVE_TYPE, target_id,))

            target_info = GoogleDriveTargetInfo(dst_node, target_node)

            proxy = None
            try:
                proxy = get_proxy("https://www.google.com/")
            except:
                proxy = None

            host = dst_type
            share = target_info.connection_info.name
            subdir = target_info.subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if  abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP
            abspath += VOL_SEP + job_name
            abspath = VOL_SEP + abspath.lstrip(VOL_SEP)
            self.__set_accesser(BackupGoogleDriveAccess(target_info, proxy))
        elif dst_type == DROPBOX_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (DROPBOX_TYPE, target_id,))

            target_info = DropboxTargetInfo(dst_node, target_node)

            proxy = None
            try:
                proxy_host, proxy_port = get_proxy_host_port(
                                            "https://www.dropbox.com/")
                proxy = proxy_host
                if proxy_port:
                    proxy += ":" + str(proxy_port)
            except:
                proxy = None

            host = dst_type
            share = target_info.connection_info.name
            subdir = target_info.subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if  abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP

            abspath += VOL_SEP + job_name
            self.__set_accesser(BackupDropboxAccess(target_info, proxy))
        elif dst_type == BOX_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (BOX_TYPE, target_id,))

            target_info = BoxTargetInfo(dst_node, target_node)

            proxy = None
            try:
                proxy = get_proxy("https://www.box.com/")
            except:
                proxy = None

            host = dst_type
            share = target_info.connection_info.name
            subdir = target_info.subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if  abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP

            abspath += VOL_SEP + job_name
            self.__set_accesser(BackupBoxAccess(target_info, proxy))
        elif dst_type == MDCLOUD_TYPE:
            target_id = dst_node.get_value(CONNECTION_ID_PATH)

            cloud_config = Request(
                fal_info.FAL_CLOUD_PATH, filter_cmd=config.SUDO_CMD).\
                request(Entry(ReqTypes.LOAD)).get_entrylist().get_entry().get()

            target_node = cloud_config.get_node(FIND_BY_ID_XPATH_TMPL %
                                            (MDCLOUD_TYPE, target_id,))

            target_info = MdCloudTargetInfo(dst_node, target_node)

            try:
                proxy_host, proxy_port = get_proxy_host_port(
                            "https://%s" %
                                (target_info.connection_info.endpoint,))
            except:
                proxy_host = None
                proxy_port = None

            host = dst_type
            share = target_info.connection_info.name + "/" + \
                                target_info.bucket
            subdir = target_info.subdir
            if subdir and subdir[0] != UNC_SEP and subdir[0] != VOL_SEP:
                subdir = VOL_SEP + subdir
            unc = \
                "\\\\%s\\%s%s" % (dst_type, share,
                                    subdir.replace(VOL_SEP, UNC_SEP))
            filesystem = None
            abspath = target_info.subdir.replace(UNC_SEP, VOL_SEP)
            if abspath:
                if  abspath[0] != VOL_SEP:
                    abspath = VOL_SEP + abspath
            else:
                abspath = VOL_SEP

            abspath += VOL_SEP + job_name
            self.__set_accesser(BackupMdCloudAccess(target_info,
                                        proxy_host, proxy_port))

        elif dst_type == WORKINGFOLDER_TYPE:
            _target_info = WorkingFolderTargetInfo(dst_node)
            _access = BackupWorkingFolderAccess(
                _target_info,
                get_proxy=get_proxy_host_port,
            )
            self.__set_accesser(_access)
            # abspath = VOL_SEP.join([_target_info.subdir, job_name])
            abspath = VOL_SEP.join(['', job_name])

        elif dst_type == HOMEBOX2_TYPE:
            _target_info = HomeBox2TargetInfo(dst_node)
            _access = BackupHomeBox2Access(
                _target_info,
                get_proxy=get_proxy_host_port,
            )
            self.__set_accesser(_access)
            abspath = VOL_SEP.join([_target_info.folder, job_name])

        generationlist_node = XmlNode(GENERATIONLIST_NAME)
        for gen_dir in self.__get_generation_dir_list(abspath):
            generationlist_node.append_child(GENERATION_NAME)
            generation_node = \
                generationlist_node.get_node_list(GENERATION_NAME)[-1]
            generation_node.append_child(ID_NAME, gen_dir)

        return generationlist_node

    def __append_generation_list_node(self, entry):

        OPT_JOBLIST_PATH = "./option/joblist"
        JOB_PATH = "backup/joblist/job"
        JOB_ID_PATH_TMPL = "./job[id='%s']"

        opt_joblist_node = entry.get().get_node(OPT_JOBLIST_PATH)

        for job_node in entry.get().get_node_list(JOB_PATH):

            job_id = job_node.get_value(ID_NAME)

            try:
                opt_job_node = \
                    opt_joblist_node.get_node(JOB_ID_PATH_TMPL % job_id)
            except NotFoundError:
                continue

            try:
                job_node.append_child(
                    self.__get_generation_list_node(opt_job_node))
            except NoEntryError:
                self.warning("no generations for job " + job_id)
                continue
            except Exception:
                self.exception("generation")
                continue  # fixme

    def __init_status(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        id = body.get_value(ID_PATH)

        self.__set_status(category, id, NOT_EXEC_STATUS)

    def __init_remain_times(self, entry):

        try:
            body = entry.get()
            category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
            id = body.get_value(ID_PATH)

            self.__set_remain_times(category, id, DISABLE_REMAIN_TIMES)
        except Exception:
            pass

    def __update_remain_times(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()

        try:
            schedule_node = body.get_node(category).get_node(SCHEDULE_PATH)
            if not schedule_node.is_enabled():
                return
            times = schedule_node.get_attr_value(TIMES_ATTR)
        except NotFoundError:
            times = "0"

        id = body.get_value(ID_PATH)
        self.__set_remain_times(category, id, times)

    def __execute_handler(self, entry):

        return self.send_entry(entry, ALBG_BACKUPS_PATH)

    def __add_handler(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        job_id = body.get_node(ID_PATH).get_text()
        if job_id and not job_id.isdigit():
            self.__init_status(entry)

        self.__update_remain_times(entry)
        return self.__add_cron_entry(entry)

    def __modify_handler(self, entry):

        self.__init_status(entry)

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        schedule_node = body.get_node(category).get_node(SCHEDULE_PATH)

        if schedule_node.is_enabled():
            self.__update_remain_times(entry)
            return self.__add_cron_entry(entry)
        else:
            return self.__delete_cron_entry(entry)

    def __delete_handler(self, entry):

        self.__init_status(entry)

        self.__init_remain_times(entry)

        return self.__delete_cron_entry(entry)

    def __query_handler(self, entry):

        try:
            for status_fname in os.listdir(config.BACKUP_STATUS_DIR):
                self.__set_status_to_xml(status_fname, entry)

            try:
                if entry.get().get_node(
                       OPTION_NAME).is_set_to_true(DETAIL_ATTR):
                    self.__append_generation_list_node(entry)
            except NotFoundError:
                pass

            entry.set_result(Results.OK)

        except Exception:
            self.exception("Query Error.")
            entry.set_result(Results.QUERY_ERROR)

        return entry

    def __apply_handler(self, entry):

        try:
            for category_node in entry.get().get_node_list(ANY_CHILD_PATH):
                for job_node in category_node.get_node_list("joblist/job"):
                    self.__set_xml_to_status(
                        category_node.get_node_name(), job_node)
            entry.set_result(Results.OK)

        except Exception:
            self.exception("Apply Error.")
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def __restore_handler(self, entry):

        return self.send_entry(entry, ALBG_BACKUPS_PATH)

    def __stop_handler(self, entry):

        return self.send_entry(entry, ALBG_BACKUPS_PATH)

    def __init__(self, option=None):
        """Constructor."""

        self.dst_listdir = None

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.EXECUTE, self.__execute_handler)
        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.RESTORE, self.__restore_handler)
        self.set_handler(ReqTypes.STOP, self.__stop_handler)

    def post_process(self, response_data):

        self.dst_listdir = None
        reload(ionas.fileaccess)
        return palcmd.PalCmd.post_process(self, response_data)


#
# Main
#

if __name__ == "__main__":
    PalBackups(option=palcmd.PalCmdOption()).handle_request()
