# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Logging utility module.

Define class for logging about backup, data copy and so on.
"""


from __future__ import with_statement

import os
import re
import time
import datetime

#from ionas.counter import (Counter, Counters)


#Backup/JOB_NAME

EPOCH_TO_BASE_NAME_FMT = "%04d%02d%02d_%02d%02d%02d"

LOG_NAME_DATE_TIME_RE = "([0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9])_*" + \
                            "([0-9][0-9][0-9][0-9][0-9][0-9])"
LOGGED_TIME_FMT = "%Y%m%d%H%M%S"

DEFAULT_LOG_MAX_COUNT = 10


class Log(object):
    """Base class for logging about backup, data copy and so on."""

    LOG_FILE_MODE = "ab"

    def _epoch_to_base_name(self, epoch):
        """Convert "epoch" to base name of log file.
           Format: YYYYMMDD_hhmmss
        """

        dinfo = datetime.datetime.fromtimestamp(epoch)

        return EPOCH_TO_BASE_NAME_FMT % (dinfo.year, dinfo.month, dinfo.day,
                                    dinfo.hour, dinfo.minute, dinfo.second)

    def _is_target_log(self, log_file, condition_list):
        """Check if "log_file" is log file corresponding to condition
           in "condition_list".
        """

        for condition in condition_list:
            match = re.match(
                LOG_NAME_DATE_TIME_RE + self.suffix_map[condition], log_file)
            if match:
                return True

        return False

    def _get_logged_time(self, log_file):
        """Return a logged epoch time for "log_file"."""

        match = re.match(LOG_NAME_DATE_TIME_RE, log_file)
        if not match:
            return 0

        logged_datetime = datetime.datetime.strptime(
                            match.group(1) + match.group(2), LOGGED_TIME_FMT)
        return int(time.mktime(logged_datetime.timetuple()))

    def _get_target_log_file_info(self, condition_list):
        """Return the log file info corresponding to condition
           in "condition_list".
        """

        log_file_info = {}
        for log_file in os.listdir(self.log_dir):
            if not self._is_target_log(log_file, condition_list):
                continue

            logged_time = self._get_logged_time(log_file)
            if not logged_time:
                continue

            if logged_time not in log_file_info:
                log_file_info[logged_time] = []

            log_file_info[logged_time].append(log_file)

        return log_file_info

    def __init__(self, log_dir, suffix_map,
            log_max_count=DEFAULT_LOG_MAX_COUNT, line_sep=os.linesep):
        """Constructor."""

        self.suffix_map = suffix_map
        self.log_dir = log_dir
        self.log_max_count = log_max_count
        self.line_sep = line_sep
        self.opened_file = None
        self.base_name = ""
        self.log_path = ""

    def __del__(self):
        """Destructor."""

        self.close()

    def __enter__(self):
        """Enter context for logging."""

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context for logging."""
        
        self.close()

    def open(self, start_epoch=None, condition=None):
        """Open log file whose name is based on "start_epoch".
           "condition" is used as key for "suffix_map".
        """

        if self.opened_file != None:
            self._close()

        self.base_name = self._epoch_to_base_name(start_epoch)
        self.log_path = os.path.join(self.log_dir,
                                self.base_name + self.suffix_map[condition])

        self.opened_file = open(os.path.join(self.log_dir, self.log_path),
                                self.LOG_FILE_MODE)
        os.chmod(os.path.join(self.log_dir, self.log_path), 0666)
        return self

    def close(self, condition=None):
        """Close log file if it is opened and rename it to name
           corresponding to "condition" if "condition" is not None,
        """

        if self.opened_file != None:
            self.opened_file.close()
            self.opened_file = None

        if condition != None:
            ended_log_path = os.path.join(self.log_dir,
                                self.base_name + self.suffix_map[condition])
            if self.log_path != ended_log_path:
                os.rename(self.log_path, ended_log_path)

    def try_delete_old_log_files(self, condition_list=None):
        """Try delete old log files corresponding to "condition" in
           "condition_list" if exists and return a list of time
           deriverd from the deleted log file name.
           If "condition_list" is None, all key in "self.suffix_map" are
           used as condition.
        """

        if condition_list == None:
            condition_list = list(self.suffix_map.keys())

        log_file_info = self._get_target_log_file_info(condition_list)
        logged_time_list = sorted(log_file_info.keys())
        log_count = len(logged_time_list)
        if log_count <= self.log_max_count:
            return []

        deleted_time_list = []
        for logged_time in logged_time_list[0:log_count - self.log_max_count]:
            deleted_time_list.append(logged_time)
            for log_file in log_file_info[logged_time]:
                os.remove(os.path.join(self.log_dir, log_file))

        return deleted_time_list

    def write(self, data):
        """Output "data" to "self.log_file" file.
           Note: A newline IS NOT outputted.
        """

        if self.opened_file != None:
            self.opened_file.write(data)

    def log(self, log_line):
        """Output "log_line" to "self.log_file" file.
           A newline is automatically outputted.
        """

        self.write(log_line + self.line_sep)


class ProcessLog(Log):
    """Class for logging process about backup, data copy and so on."""

    PROCESSING = "processing"
    SUCCESS = "success"
    FAILED = "failed"

    LOG_SUFFIX_MAP = {
            PROCESSING: "_processing.log",
            SUCCESS: ".log",
            FAILED: "_failed.log",
        }

    def __init__(self, log_dir, log_max_count=DEFAULT_LOG_MAX_COUNT,
            line_sep=os.linesep):
        """Constructor."""

        Log.__init__(self, log_dir, self.LOG_SUFFIX_MAP, log_max_count,
                line_sep)

    def open(self, start_epoch=None):
        """Open log file whose name is based on "start_epoch".
           "condition" is used as key for "suffix_map".
        """

        return Log.open(self, start_epoch, self.PROCESSING)


class SummaryLog(Log):
    """Class for logging about backup, data copy and so on."""
