#!/usr/bin/env python
# -*- coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import division

import os

from django.utils.translation import ugettext as _
from django import forms
from django.forms.formsets import formset_factory
from project.view import FuncView as _View

from .fal.sync import (
    obtain_syncconfs,
    create_syncconf,
    update_syncconf,
    delete_syncconf,
)
from . import fields
from . import storages


ENTRY_MAX_NUM = 10

STORAGES = storages.storages


def _get_entries():
    _entries = []
    for _syncconf in obtain_syncconfs():
        _entry = _syncconf
        _entry['pk'] = _entry['id']
        _entries.append(_entry)
    return _entries


def _get_entry(pk):
    for entry in _get_entries():
        if entry['pk'] == pk:
            return entry
    return None


class _EntryForm(forms.Form):

    def __init__(self, *args, **kwargs):
        super(_EntryForm, self).__init__(*args, **kwargs)
        self.fields['pk'] = forms.CharField(
            widget=forms.widgets.HiddenInput,
        )
        self.fields['selected'] = forms.BooleanField(
            widget=forms.widgets.CheckboxInput,
            required=False,
        )
EntryFormSet = formset_factory(
    _EntryForm,
    extra=0,
    max_num=ENTRY_MAX_NUM,
    can_delete=False,
)


def _get_entries_and_formset(data=None):
    entries = _get_entries()
    if data:
        formset = EntryFormSet(data=data)
    else:
        initials = [{
            'pk': e['pk'],
            'selected': False,
        } for e in entries]
        formset = EntryFormSet(initial=initials)
    for entry, form in zip(entries, formset):
        entry['form'] = form
    return entries, formset


class EntriesView(_View):

    views = ['save', 'delete']

    def view(self, request):
        entries, formset = _get_entries_and_formset()

        template = os.path.join(self.TEMPLATES, 'entries.html')
        dictionary = {
            'entries': entries,
            'formset': formset,
        }
        return {
            self.KEY_TEMPLATE: template,
            self.KEY_TEMPLATE_DIC: dictionary,
        }

    def view_delete(self, request):
        entries, formset = _get_entries_and_formset(data=request.POST)

        with self.processing() as proc:
            is_succeeded = True
            if formset.is_valid():
                for cleaned_data in formset.cleaned_data:
                    if cleaned_data['selected']:
                        pk = cleaned_data['pk']
                        ret = delete_syncconf(pk)
                        is_succeeded = is_succeeded and ret['is_succeeded']
                        self.set_message(ret['message'])
            proc.result(is_succeeded)
        entries, formset = _get_entries_and_formset()

        template = os.path.join(self.TEMPLATES, 'entries.html')
        dictionary = {
            'entries': entries,
            'formset': formset,
        }
        return {
            self.KEY_TEMPLATE: template,
            self.KEY_TEMPLATE_DIC: dictionary,
        }


class MainForm(forms.Form):

    name = forms.CharField(label=u"ジョブ名")
    direction = fields.DirectionField()
    interval = fields.IntervalField()
    maxfilesize = fields.MaxFileSizeField(required=False)

    def __init__(self, *args, **kwargs):
        kwargs['prefix'] = kwargs.get('prefix', 'main')
        super(MainForm, self).__init__(*args, **kwargs)


class _IgnoreForm(forms.Form):
    ignore_pattern = fields.IgnorePatternField()
    ignore_type = fields.IgnoreTypeField()
_IgnoreFormSet = formset_factory(
    _IgnoreForm,
    extra=0,
    max_num=10,
    can_delete=True,
)
class IgnoreFormSet(_IgnoreFormSet):
    def __init__(self, *args, **kwargs):
        kwargs['prefix'] = kwargs.get('prefix', 'ignore')
        super(IgnoreFormSet, self).__init__(*args, **kwargs)


class _TargetForm(forms.Form):

    _STORAGES = {}
    for s in STORAGES:
        _STORAGES[s['pk']] = {
            'pk': s['pk'],
            'Form': s['Form'],
            'STYLE_HTML': s['STYLE_HTML'],
            'SCRIPT_HTML': s['SCRIPT_HTML'],
            'choice': (s['pk'], s['name']),
            'order': s['order'],
        }

    def __init__(
        self,
        prefix=None,
        initial=None,
        data=None,
        files=None,
        **kwargs):
        super(_TargetForm, self).__init__(
            prefix=prefix,
            initial=initial,
            data=data,
            files=files,
            **kwargs)
        self.fields['spk'] = forms.ChoiceField(
            label=u"対象種別",
            choices=[v['choice'] for v in sorted(self.storages.values(), key=lambda v: v['order'])],
        )
        self.subforms = {}
        for _pk, _parameters in self._STORAGES.iteritems():
            _prefix = '{0}-{1}'.format(prefix, _pk)
            _initial = None
            _data = None
            _files = None
            if isinstance(initial, dict) and initial.get('spk') == _pk:
                _initial = initial
            if self['spk'].value() == _pk:
                _data = data
                _files = files
            _Form = _parameters['Form']
            _form = _Form(
                prefix=_prefix,
                initial=_initial,
                data=_data,
                files=_files,
            )
            _form.STYLE_HTML = _parameters['STYLE_HTML']  # TODO
            _form.SCRIPT_HTML = _parameters['SCRIPT_HTML']  # TODO
            self.subforms[_pk] = _form

    def _get_storages(self):
        return self._STORAGES
    storages = property(_get_storages)

    def _get_errors(self):
        self._errors = super(_TargetForm, self)._get_errors()
        cleaned_data = getattr(self, 'cleaned_data', None)
        for subform in self.subforms.values():
            self._errors.update(subform.errors)
            if cleaned_data is not None:
                cleaned_data.update(getattr(subform, 'cleaned_data', {}))
        return self._errors
    errors = property(_get_errors)


_TargetFormset = formset_factory(
    _TargetForm,
    extra=0,
    max_num=2,
    can_delete=False,
)


class TargetFormset(_TargetFormset):

    def __init__(self, *args, **kwargs):
        kwargs['prefix'] = kwargs.get('prefix', 'target')
        super(TargetFormset, self).__init__(*args, **kwargs)


class EntryView(_View):

    views = ['save']

    def view(self, request, pk=None):
        entry = _get_entry(pk)
        if pk is not None and entry is None:
            self.set_message(u"対象のジョブが存在しません。")
        initial = self._get_initial(entry)

        mainform = MainForm(initial=initial['mainform'])
        ignoreformset = IgnoreFormSet(initial=initial['ignoreformset'])
        targetformset = TargetFormset(initial=initial['targetformset'])

        template = os.path.join(self.TEMPLATES, 'entry.html')
        dictionary = {
            'mainform': mainform,
            'ignoreformset': ignoreformset,
            'targetformset': targetformset,

            'TARGETS_HTML': os.path.join(self.TEMPLATES, '_targets.html'),
            'TARGET_HTML': os.path.join(self.TEMPLATES, '_target.html'),
        }
        return {
            self.KEY_TEMPLATE: template,
            self.KEY_TEMPLATE_DIC: dictionary,
        }

    def view_save(self, request, pk=None):
        entry = _get_entry(pk)
        if pk is not None and entry is None:
            self.set_message(u"対象のジョブが存在しません。")

        mainform = MainForm(data=request.POST)
        ignoreformset = IgnoreFormSet(data=request.POST)
        targetformset = TargetFormset(data=request.POST)

        with self.processing() as proc:
            if mainform.is_valid() and ignoreformset.is_valid() and targetformset.is_valid():
                pass
                parameters = {
                    'name': mainform.cleaned_data['name'],
                    'src': targetformset.cleaned_data[0],
                    'dst': targetformset.cleaned_data[1],
                    'ignores': [{
                        'type': data['ignore_type'],
                        'pattern': data['ignore_pattern']
                    } for data in ignoreformset.cleaned_data],
                    'direction': mainform.cleaned_data['direction'],
                    'interval': mainform.cleaned_data['interval'],
                    'maxfilesize': mainform.cleaned_data['maxfilesize'],
                }
                if pk is None:
                    save_func = create_syncconf
                else:
                    save_func = update_syncconf
                    parameters['id'] = pk
                ret = save_func(parameters)
                proc.result(ret['is_succeeded'])
                self.set_message(ret['message'])

        template = os.path.join(self.TEMPLATES, 'entry.html')
        dictionary = {
            'mainform': mainform,
            'ignoreformset': ignoreformset,
            'targetformset': targetformset,

            'STORAGES': STORAGES,
            'TARGETS_HTML': os.path.join(self.TEMPLATES, '_targets.html'),
            'TARGET_HTML': os.path.join(self.TEMPLATES, '_target.html'),
        }
        return {
            self.KEY_TEMPLATE: template,
            self.KEY_TEMPLATE_DIC: dictionary,
        }

    def _get_initial(self, entry):
        try:
            return {
                'mainform': {
                    'name': entry['name'],
                    'direction': entry['direction'],
                    'interval': entry['interval'],
                    'maxfilesize': entry['maxfilesize'],
                },
                'ignoreformset': [{
                    'ignore_type': _ignore['type'],
                    'ignore_pattern': _ignore['pattern'],
                } for _ignore in entry['ignores']],
                'targetformset': [
                    entry['src'],
                    entry['dst'],
                ],
            }
        except:
            return {
                'mainform': {},
                'ignoreformset': [],
                'targetformset': [
                    {},
                    {},
                ],
            }
